#ifndef _GFX_ABSTRACT_CONTROLLER_H
#define _GFX_ABSTRACT_CONTROLLER_H

#include <stdbool.h>

/**
 * @file gfx-abstract-controller.h
 *
 * Abstraction de contrôleur d'affichage graphique monochrome.
 * Définit le contrat à implémenter pour un contrôleur précis.
 */

#include <stdbool.h>

#include "gfx-module.h"

typedef enum {
	ORIENTATION_LEFT_TO_RIGHT = 0,
	ORIENTATION_RIGHT_TO_LEFT = 1,
} GfxHorizontalOrientation;

typedef enum {
	ORIENTATION_TOP_TO_BOTTOM = 0,
	ORIENTATION_BOTTOM_TO_TOP = 8,
} GfxVerticalOrientation;

typedef enum {
	MODE_SLEEP = 0,
	MODE_ACTIVE = 1,
} GfxDisplayMode;

/**
 * Cette structure regroupe les fonctions que tout contrôleur doit implémenter
 * pour qu'on soit en mesure d'utiliser un module graphique monochrome.
 */
struct GfxController {
	/**
	 * Initialise le contrôleur.
	 */
	void (*initialise)(GfxModule *module);

	/**
	 * Efface l'écran et positionne le curseur en haut à gauche.
	 */
	void (*clear)(GfxModule *module);

	/**
	 * Configure les adresses de rangée et de colonne en vue d'une prochaine lecture ou écriture.
	 */
	void (*setAddress)(GfxModule *module, uint8_t xAddress, uint8_t yAddress);

	/**
	 * Ecrit un bloc de données dans la mémoire du module OLED.
	 * A utiliser après setAddress().
	 *
	 * @param data Adresse du bloc à écrire.
	 * @param count Nombre d'octets à écrire.
	 */
	void (*writeData)(GfxModule *module, const uint8_t *data, int count);

	/**
	 * @param contrast Valeur du contraste (par défaut : 128).
	 */
	void (*setContrast)(GfxModule *module, uint8_t contrast);

	/**
	 * Définit l'orientation de l'affichage.
	 *
	 * @param horizontal
	 * @param vertical
	 */
	void (*setOrientation)(GfxModule *module, GfxHorizontalOrientation horizontal, GfxVerticalOrientation vertical);

	/**
	 * Définit si les couleurs de l'affichage sont inversées.
	 *
	 * @param inverse true = affichage noir sur fond de couleur,
	 * false = affichage couleur sur fond noir (par défaut).
	 */
	void (*setInverseMode)(GfxModule *module, bool inverse);

	/**
	 * Définit le mode de fonctionnement du module.
	 * Par défaut, le module est en sleep mode.
	 *
	 * @param module
	 * @param mode
	 */
	void (*setMode)(GfxModule *module, GfxDisplayMode mode);
};

#endif // _GFX_ABSTRACT_CONTROLLER_H
