#include "project-defs.h"
#include "gfx-module.h"

#include <stdlib.h>
#include <string.h>
#include "gfx-abstract-controller.h"

static void beginOperation(GfxModule *module) {
	if (!module->internals.batchStarted) {
		module->internals.batchStarted = true;
		module->internals.minExtentX = module->display.width;
		module->internals.minExtentY = module->display.height;
		module->internals.maxExtentX = 0;
		module->internals.maxExtentY = 0;
	}
}

static void endOperation(GfxModule *module) {
	if (module->internals.autoUpdate) {
		gfxUpdateDisplay(module);
	}
}

static void drawPoint(GfxModule *module, uint8_t x, uint8_t y, GfxColour colour) {
	// Chaque octet de la mémoire tampon contient 8 pixels verticaux
	// consécutifs rangés dans l'ordre croissant des poids des bits.
	unsigned int byteOffset = y / 8 * module->display.width + x;

	uint8_t byte = module->buffer[byteOffset];
	uint8_t bitMask = 1 << (y % 8);

	switch (colour) {
	case COLOUR_OFF:
		byte &= ~bitMask;
		break;
	case COLOUR_ON:
		byte |= bitMask;
		break;
	case COLOUR_INVERSE:
		byte ^= bitMask;
		break;
	}

	module->buffer[byteOffset] = byte;

	if (x < module->internals.minExtentX) {
		module->internals.minExtentX = x;
	}

	if (x > module->internals.maxExtentX) {
		module->internals.maxExtentX = x;
	}

	if (y < module->internals.minExtentY) {
		module->internals.minExtentY = y;
	}

	if (y > module->internals.maxExtentY) {
		module->internals.maxExtentY = y;
	}
}

void gfxEnableAutoUpdate(GfxModule *module) {
	module->internals.autoUpdate = true;
}

void gfxDisableAutoUpdate(GfxModule *module) {
	gfxUpdateDisplay(module);
	module->internals.autoUpdate = false;
}

void gfxUpdateDisplay(GfxModule *module) {
	if (module->internals.batchStarted) {
		module->internals.batchStarted = false;
		// On aligne les coordonnées verticales sur des frontières de 8 bits
		// et les coordonnées horizontales sur des frontières de 16 bits afin
		// de s'adapter aux contraintes de tous les contrôleurs.
		uint8_t yStart = module->internals.minExtentY & ~0x07;
		uint8_t yEnd = module->internals.maxExtentY | 0x07;
		uint8_t byteCount = (module->internals.maxExtentX | 0x0f) - (module->internals.minExtentX & ~0x0f) + 1;

		for (uint8_t y = yStart; y <= yEnd; y += 8) {
			module->controller->setAddress(module, module->internals.minExtentX, y);
			unsigned int byteOffset = (y / 8) * module->display.width + module->internals.minExtentX;
			// Un octet du buffer contient 8 pixels verticaux consécutifs
			// rangés dans l'ordre croissant des poids des bits. writeData()
			// reçoit autant d'octets que de valeurs différentes de x.
			module->controller->writeData(module, module->buffer + byteOffset, byteCount);
		}
	}
}

void gfxInitialise(GfxModule *module) {
	commLinkInitialise(&module->link);
	module->controller->initialise(module);
	module->internals.autoUpdate = false;
	module->internals.batchStarted = false;
	// Le "+ 7" permet de gérer l'arrondi à l'entier supérieur lorsque
	// module->display.height n'est pas un multiple de 8.
	module->internals.nrPages = (module->display.height + 7) / 8;
	memset(module->buffer, 0, module->display.width * module->internals.nrPages);
}

void gfxClear(GfxModule *module) {
	beginOperation(module);
	memset(module->buffer, 0, module->display.width * module->internals.nrPages);
	module->internals.minExtentX = 0;
	module->internals.minExtentY = 0;
	module->internals.maxExtentX = module->display.width - 1;
	module->internals.maxExtentY = module->display.height - 1;
	endOperation(module);
}

void gfxPoint(GfxModule *module, uint8_t x, uint8_t y, GfxColour colour) {
	beginOperation(module);
	drawPoint(module, x, y, colour);
	endOperation(module);
}

// On utilise l'algorithme de Bresenham :
// https://fr.wikipedia.org/wiki/Algorithme_de_trac%C3%A9_de_segment_de_Bresenham
void gfxLine(GfxModule *module, uint8_t fromX, uint8_t fromY, uint8_t toX, uint8_t toY, GfxColour colour) {
	beginOperation(module);

	if (fromX == toX) {
		if (fromY > toY) {
			uint8_t n = toY;
			toY = fromY;
			fromY = n;
		}

		for (uint8_t y = fromY; y <= toY; y++) {
			drawPoint(module, fromX, y, colour);
		}
	} else if (fromY == toY) {
		if (fromX > toX) {
			uint8_t n = toX;
			toX = fromX;
			fromX = n;
		}

		for (uint8_t x = fromX; x <= toX; x++) {
			drawPoint(module, x, fromY, colour);
		}
	} else {
		int dx = abs(toX - fromX);
		int sx = fromX < toX ? 1 : -1;
		int dy = -abs(toY - fromY);
		int sy = fromY < toY ? 1 : -1;
		int err = dx + dy;
		int e2;
		int x = fromX;
		int y = fromY;

		while (1) {
			drawPoint(module, x, y, colour);
			e2 = 2 * err;

			if (e2 >= dy) {
				if (x == toX)
					break;

				err += dy;
				x += sx;
			}

			if (e2 <= dx) {
				if (y == toY)
					break;

				err += dx;
				y += sy;
			}
		}
	}

	endOperation(module);
}

void gfxImage(GfxModule *module,
	uint8_t imageWidth, uint8_t imageHeight, uint8_t *imageBits,
	GfxAlignment alignment, uint8_t positionX, uint8_t positionY,
	GfxColour colour) {
	beginOperation(module);
	uint8_t offsetX = 0;
	uint8_t offsetY = 0;

	switch (alignment) {
	case ALIGN_TOP_LEFT:
	case ALIGN_BOTTOM_LEFT:
	case ALIGN_MIDDLE_LEFT:
		offsetX = 0;
		break;
	case ALIGN_TOP_RIGHT:
	case ALIGN_BOTTOM_RIGHT:
	case ALIGN_MIDDLE_RIGHT:
		offsetX = (module->display.width > imageWidth) ? (module->display.width - imageWidth) : 0;
		break;
	case ALIGN_TOP_CENTER:
	case ALIGN_BOTTOM_CENTER:
	case ALIGN_MIDDLE_CENTER:
		offsetX = (module->display.width > imageWidth) ? ((module->display.width - imageWidth) / 2) : 0;
		break;
	case ALIGN_CUSTOM_OFFSET:
		offsetX = positionX;
		break;
	}

	switch (alignment) {
	case ALIGN_TOP_LEFT:
	case ALIGN_TOP_RIGHT:
	case ALIGN_TOP_CENTER:
		offsetY = 0;
		break;
	case ALIGN_BOTTOM_LEFT:
	case ALIGN_BOTTOM_RIGHT:
	case ALIGN_BOTTOM_CENTER:
		offsetY = (module->display.height > imageHeight) ? (module->display.height - imageHeight) : 0;
		break;
	case ALIGN_MIDDLE_LEFT:
	case ALIGN_MIDDLE_RIGHT:
	case ALIGN_MIDDLE_CENTER:
		offsetY = (module->display.height > imageHeight) ? ((module->display.height - imageHeight) / 2) : 0;
		break;
	case ALIGN_CUSTOM_OFFSET:
		offsetY = positionY;
		break;
	}

	uint8_t bytesPerImageRow = (imageWidth + 7) / 8;

	for (uint8_t row = 0; row < imageHeight; row++) {
		for (uint8_t col = 0; col < bytesPerImageRow; col++) {
			uint8_t byte = imageBits[row * bytesPerImageRow + col];

			for (uint8_t bit = 0; bit < 8; bit++) {
				// Format XBM : le pixel le plus à gauche est représenté par
				// le bit de poids le plus faible.
				if (byte & 1) {
					drawPoint(module, offsetX + col * 8 + bit, offsetY + row, colour);
				}

				byte >>= 1;
			}
		}
	}

	uint8_t xMin = offsetX;
	uint8_t yMin = offsetY;
	uint8_t xMax = xMin + imageWidth - 1;
	uint8_t yMax = yMin + imageHeight - 1;

	if (xMin < module->internals.minExtentX) {
		module->internals.minExtentX = xMin;
	}

	if (xMax > module->internals.maxExtentX) {
		module->internals.maxExtentX = xMax;
	}

	if (yMin < module->internals.minExtentY) {
		module->internals.minExtentY = yMin;
	}

	if (yMax > module->internals.maxExtentY) {
		module->internals.maxExtentY = yMax;
	}

	endOperation(module);
}
