#ifndef _GFX_MODULE_H
#define _GFX_MODULE_H

/**
 * @file gfx-module.h
 *
 * Définitions relatives au module d'affichage graphique monochrome dans son
 * ensemble et fonctions de haut niveau pour l'utiliser facilement.
 */

#include <stdbool.h>

#include "comm-link-abstract.h"
#include "gfx-display.h"

/**
 * "Couleurs" utilisées pour le tracé des points par les fonctions de dessin.
 */
typedef enum {
	COLOUR_OFF = 0,/**< Ecrit un 0 dans la mémoire du contrôleur. */
	COLOUR_ON,     /**< Ecrit un 1 dans la mémoire du contrôleur. */
	COLOUR_INVERSE,/**< Complémente le contenu de la mémoire du contrôleur. */
} GfxColour;

struct GfxController;

/**
 * Etat interne de GFXModule, ne pas modifier depuis votre code.
 */
typedef struct {
	bool autoUpdate;
	bool batchStarted;
	uint8_t minExtentX;
	uint8_t minExtentY;
	uint8_t maxExtentX;
	uint8_t maxExtentY;
	uint8_t nrPages;
} GfxModuleInternals;

/**
 * Cette structure contient la description de tous les composants d'un
 * module d'affichage graphique monochrome. Pour pouvoir utiliser un module,
 * il faut initialiser chaque membre de cette structure avec les données
 * de configuration et les blocs de fonctions correspondant aux éléments
 * de la data sheet du module.
 */
typedef struct {
	// Caractéristiques de l'écran
	const GfxDisplay display;
	// Contrôleur
	const struct GfxController *controller;
	// Interface de communication
	const CommLink link;
	/**
	 * Buffer d'affichage permettant de limiter les communications avec le module.
	 * C'est à vous de l'allouer de manière statique. Exemple :
	 *
#include "oled-i2c-link.h"
#include "oled-sh1106-controller.h"
#include "oled-module.h"

#define DISPLAY_WIDTH 128
#define DISPLAY_HEIGHT 64

static const OLEDI2CConfig OLED_I2C_CONFIG = { .i2cAddress = 0x3c, };

// Le "+ 7" permet de gérer l'arrondi à l'entier supérieur lorsque
// DISPLAY_HEIGHT n'est pas un multiple de 8.
static uint8_t oledBuffer[DISPLAY_WIDTH * (DISPLAY_HEIGHT + 7) / 8];

OLEDModule sh1106 = {
	.display = { .width = DISPLAY_WIDTH, .height = DISPLAY_HEIGHT, },
	.controller = &OLED_SH1106_CONTROLLER,
	.link = { .implementation = &OLED_I2C_LINK_IMPL, .configuration = &OLED_I2C_CONFIG, },
	.buffer = oledBuffer,
};
	 */
	uint8_t *buffer;

	// Etat interne de OLEDModule, ne pas modifier depuis votre code.
	GfxModuleInternals internals;
} GfxModule;

/**
 * Active la mise à jour automatique de l'affichage en fin d'exécution
 * de chaque fonction.
 *
 * @param module
 */
void gfxEnableAutoUpdate(GfxModule *module);

/**
 * Désactive la mise à jour automatique de l'affichage en fin d'exécution
 * de chaque fonction. C'est la valeur par défaut car il est plus judicieux
 * de ne mettre à jour l'affichage qu'une fois qu'on a terminé toutes les
 * opérations de dessin.
 *
 * @param module
 */
void gfxDisableAutoUpdate(GfxModule *module);

/**
 * Fonction à appeler pour mettre à jour l'affichage lorsque la mise à jour
 * automatique est désactivée. Si elle est activée, appeler cette fonction
 * n'a aucun effet (ça a déjà été fait automatiquement).
 *
 * @param module
 */
void gfxUpdateDisplay(GfxModule *module);

/**
 * Initialisation du module.
 *
 * @param module Pointeur sur la structure de description du module OLED.
 */
void gfxInitialise(GfxModule *module);

/**
 * Efface complètement l'affichage.
 *
 * @param module Pointeur sur la structure de description du module OLED.
 */
void gfxClear(GfxModule *module);

/**
 * Trace un point aux coordonnées indiquées.
 *
 * @param module
 * @param x
 * @param y
 * @param colour
 */
void gfxPoint(GfxModule *module, uint8_t x, uint8_t y, GfxColour colour);

/**
 * Trace un segment entre les 2 points indiqués.
 * Les coordonnées commencent à 0 et croissent vers la droite et le bas.
 *
 * @param module Pointeur sur la structure de description du module OLED.
 * @param fromX Point de départ (colonne).
 * @param fromY Point de départ (ligne).
 * @param toX Point d'arrivée (colonne).
 * @param toY Point d'arrivée (ligne).
 * @param colour
 */
void gfxLine(GfxModule *module, uint8_t fromX, uint8_t fromY, uint8_t toX, uint8_t toY, GfxColour colour);

/**
 * Types d'alignement de l'image sur l'écran.
 */
typedef enum {
	ALIGN_TOP_LEFT,     /**< L'image est placée dans le coin haut gauche de l'écran. */
	ALIGN_TOP_CENTER,   /**< L'image est centrée horizontalement et alignée sur le bord supérieur de l'écran. */
	ALIGN_TOP_RIGHT,    /**< L'image est placée dans le coin haut droit de l'écran. */
	ALIGN_BOTTOM_LEFT,  /**< L'image est placée dans le coin bas gauche de l'écran. */
	ALIGN_BOTTOM_CENTER,/**< L'image est centrée horizontalement et alignée sur le bord inférieur de l'écran. */
	ALIGN_BOTTOM_RIGHT, /**< L'image est placée dans le coin bas droit de l'écran. */
	ALIGN_MIDDLE_LEFT,  /**< L'image est centrée verticalement et alignée sur le bord gauche de l'écran. */
	ALIGN_MIDDLE_RIGHT, /**< L'image est centrée verticalement et alignée sur le bord droit de l'écran. */
	ALIGN_MIDDLE_CENTER,/**< L'image est centrée verticalement et horizontalement. */
	ALIGN_CUSTOM_OFFSET,/**< Le coin supérieur gauche de l'image est placé à (positionX, positionY). */
} GfxAlignment;

/**
 * Affiche une image au format XBM.
 * Remarque : les bits de l'image qui sont à 0 sont transparents.
 * Ceux qui sont à 1 sont tracés selon la valeur de "colour".
 *
 * @param module
 * @param imageWidth Utiliser la constante définie dans le fichier .xbm.
 * @param imageHeight Utiliser la constante définie dans le fichier .xbm.
 * @param imageBits Utiliser la constante définie dans le fichier .xbm.
 * @param alignment Type d'alignement de l'image.
 * @param positionX Utilisé seulement si alignment == ALIGN_CUSTOM_OFFSET.
 * @param positionY Utilisé seulement si alignment == ALIGN_CUSTOM_OFFSET.
 * @param colour
 */
void gfxImage(
	GfxModule *module,
	uint8_t imageWidth, uint8_t imageHeight, uint8_t *imageBits,
	GfxAlignment alignment, uint8_t positionX, uint8_t positionY,
	GfxColour colour
);

#endif // _GFX_MODULE_H
