#include "project-defs.h"
#define _GFX_SH1106_CONTROLLER_C
#include "gfx-sh1106-controller.h"
#include <string.h>

#define PAGE_COUNT 8
#define COLUMN_COUNT 132

static void setAddress(GfxModule *module, uint8_t xAddress, uint8_t yAddress) {
	uint8_t buffer[3];

	buffer[0] = 0xb0 | (yAddress >> 3);
	// L'écran fait 128 pixels de large, mais le SH1106 gère jusqu'à 132
	// colonnes. Or, il se trouve que les 128 pixels de l'écran sont centrés
	// dans ces 132 colonnes, ce qui fait que la coordonnée x = 0 commence
	// à la colonne 2 dans la mémoire du SH1106.
	// Cette information ne se trouve pas dans la documentation disponible
	// (elle figure dans celle du module et nous n'avons que celle du
	// contrôleur), le problème s'est donc révélé au debugging.
	xAddress += 2;
	buffer[1] = 0x10 | (xAddress >> 4);
	buffer[2] = 0x00 | (xAddress & 0x0f);
	commLinkWrite(&module->link, TYPE_COMMAND, buffer, sizeof(buffer));
}

static void writeData(GfxModule *module, const uint8_t *data, int count) {
	commLinkWrite(&module->link, TYPE_DATA, data, count);
}

static void clear(GfxModule *module) {
	uint8_t buffer[COLUMN_COUNT];

	memset(buffer, 0, sizeof(buffer));

	for (int page = 0; page < PAGE_COUNT; page++) {
		setAddress(module, page, 0);
		writeData(module, buffer, sizeof(buffer));
	}
}

static void setMode(GfxModule *module, GfxDisplayMode mode) {
	uint8_t buffer = 0xae | mode;
	commLinkWrite(&module->link, TYPE_COMMAND, &buffer, 1);
}

static void setOrientation(GfxModule *module, GfxHorizontalOrientation horizontal, GfxVerticalOrientation vertical) {
	uint8_t buffer[2];

	buffer[0] = 0xa0 | horizontal;
	buffer[1] = 0xc0 | vertical;
	commLinkWrite(&module->link, TYPE_COMMAND, buffer, sizeof(buffer));
}

static void initialise(GfxModule *module) {
	// Petit délai pour être sûr que le SH1106 ait fini son reset.
	Delay_Us(100);
	// Voir page 33 de la data sheet du SH1106.
	clear(module);
	// L'écran est monté à l'envers si le connecteur est en haut.
	setOrientation(module, ORIENTATION_LEFT_TO_RIGHT, ORIENTATION_BOTTOM_TO_TOP);
	setMode(module, MODE_ACTIVE);
	Delay_Ms(100);
}

static void setContrast(GfxModule *module, uint8_t contrast) {
	uint8_t buffer[2];

	buffer[0] = 0x81;
	buffer[1] = contrast;
	commLinkWrite(&module->link, TYPE_COMMAND, buffer, sizeof(buffer));
}

static void setInverseMode(GfxModule *module, bool inverse) {
	uint8_t buffer = 0xa6 | (inverse ? 1 : 0);
	commLinkWrite(&module->link, TYPE_COMMAND, &buffer, 1);
}

const struct GfxController GFX_SH1106_CONTROLLER = {
	.initialise = initialise,
	.clear = clear,
	.writeData = writeData,
	.setAddress = setAddress,
	.setContrast = setContrast,
	.setOrientation = setOrientation,
	.setInverseMode = setInverseMode,
	.setMode = setMode,
};
