#include "project-defs.h"
#include <stdbool.h>
#include <debug.h>

#include "comm-link-i2c.h"
#include "gfx-sh1106-controller.h"
#include "gfx-module.h"
#include "smiley.xbm"

#define DISPLAY_WIDTH 128
#define DISPLAY_HEIGHT 64

static const CommLinkI2CConfig COMM_I2C_CONFIG = { .i2cAddress = 0x3c, };

// Le "+ 7" permet de gérer l'arrondi à l'entier supérieur lorsque
// DISPLAY_HEIGHT n'est pas un multiple de 8.
static uint8_t oledBuffer[DISPLAY_WIDTH * (DISPLAY_HEIGHT + 7) / 8];

GfxModule gfxModule = {
	// L'affichage comporte 64 lignes de 128 pixels.
	.display = { .width = DISPLAY_WIDTH, .height = DISPLAY_HEIGHT, },
	// Le contrôleur est un SH1106 ou compatible.
	.controller = &GFX_SH1106_CONTROLLER,
	// La liaison entre micro-contrôleur et module OLED est un bus I2C.
	.link = { .implementation = &COMM_LINK_I2C_IMPL, .configuration = &COMM_I2C_CONFIG, },
	// On a alloué un buffer d'affichage de manière statique.
	.buffer = oledBuffer,
};

// Le SH1106 supporte jusqu'à 400kHz (data sheet page 1).
#define I2C_CLOCK_FREQ 400000
// On assigne l'adresse I2C 1 au CH32V003.
#define I2C_HOST_ADDR 1

// Pause entre chaque motif de test, en millisecondes
#define PAUSE_DURATION 1000

int main() {
	Delay_Init();
	USART_Printf_Init(115200);

	/*
	 * IMPORTANT : on laisse l'initialisation de l'interface I2C ici car
	 * il pourrait y avoir plusieurs périphériques I2C connectés au bus.
	 * Cela signifie que cette initialisation n'a pas sa place dans le
	 * code dédié au lien I2C avec le module d'affichage graphique.
	 */

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_AFIO, ENABLE);

	GPIO_InitTypeDef gpioInit = { 0 };
	// PC1 = SDA
	// PC2 = SCL
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOC, ENABLE);
	gpioInit.GPIO_Pin = GPIO_Pin_1 | GPIO_Pin_2;
	gpioInit.GPIO_Mode = GPIO_Mode_AF_OD;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIOC, &gpioInit);

	RCC_APB1PeriphClockCmd(RCC_APB1Periph_I2C1, ENABLE);
	I2C_InitTypeDef i2cInit = { 0 };
	i2cInit.I2C_ClockSpeed = I2C_CLOCK_FREQ;
	i2cInit.I2C_Mode = I2C_Mode_I2C;
	i2cInit.I2C_DutyCycle = I2C_DutyCycle_2;
	i2cInit.I2C_OwnAddress1 = I2C_HOST_ADDR << 1;
	i2cInit.I2C_Ack = I2C_Ack_Enable;
	i2cInit.I2C_AcknowledgedAddress = I2C_AcknowledgedAddress_7bit;
	I2C_Init(I2C1, &i2cInit);
	I2C_Cmd(I2C1, ENABLE);

	// On initialise le module LCD
	gfxInitialise(&gfxModule);

	while (1) {
		// Test d'affichage d'image
		gfxClear(&gfxModule);
		gfxImage(&gfxModule, smiley_width, smiley_height, smiley_bits, ALIGN_MIDDLE_CENTER, 0, 0, COLOUR_ON);
		gfxUpdateDisplay(&gfxModule);
		Delay_Ms(PAUSE_DURATION);

		// Test de lignes verticales
		gfxClear(&gfxModule);

		for (int col = 0; col < gfxModule.display.width; col += 8) {
			gfxLine(&gfxModule, col, 0, col, gfxModule.display.height - 1, COLOUR_ON);
		}

		gfxUpdateDisplay(&gfxModule);
		Delay_Ms(PAUSE_DURATION);

		// Test de lignes horizontales
		gfxClear(&gfxModule);

		for (int lin = 0; lin < gfxModule.display.height; lin += 8) {
			gfxLine(&gfxModule, 0, lin, gfxModule.display.width - 1, lin, COLOUR_ON);
		}

		gfxUpdateDisplay(&gfxModule);
		Delay_Ms(PAUSE_DURATION);

		// Test de diagonales
		gfxClear(&gfxModule);

		for (int col = 0; col < (gfxModule.display.width - 32); col += 8) {
			gfxLine(&gfxModule, col, 0, col + 31, gfxModule.display.height - 1, COLOUR_ON);
		}

		gfxUpdateDisplay(&gfxModule);
		Delay_Ms(PAUSE_DURATION);
	}
}
